@@ -106,10 +106,8 @@ int INTEGRITY_data_sectors(struct crypt_device *cd,
 	return 0;
 }
 
-int INTEGRITY_key_size(struct crypt_device *cd)
+int INTEGRITY_key_size(struct crypt_device *cd, const char *integrity)
 {
-	const char *integrity = crypt_get_integrity(cd);
-
 	if (!integrity)
 		return 0;
 
@@ -218,7 +216,7 @@ int INTEGRITY_activate(struct crypt_device *cd,
 
 	r = dm_create_device(cd, name, "INTEGRITY", &dmdi, 0);
 	if (r < 0 && (dm_flags(DM_INTEGRITY, &dmi_flags) || !(dmi_flags & DM_INTEGRITY_SUPPORTED))) {
-		log_err(cd, _("Kernel doesn't support dm-integrity mapping.\n"));
+		log_err(cd, _("Kernel doesn't support dm-integrity mapping."));
 		return -ENOTSUP;
 	}
 
@@ -269,13 +267,19 @@ int INTEGRITY_format(struct crypt_device *cd,
 
 	r = device_block_adjust(cd, dmdi.data_device, DEV_EXCL, dmdi.u.integrity.offset, NULL, NULL);
 	if (r < 0 && (dm_flags(DM_INTEGRITY, &dmi_flags) || !(dmi_flags & DM_INTEGRITY_SUPPORTED))) {
-		log_err(cd, _("Kernel doesn't support dm-integrity mapping.\n"));
+		log_err(cd, _("Kernel doesn't support dm-integrity mapping."));
 		return -ENOTSUP;
 	}
 	if (r)
 		return r;
 
+	/* There is no data area, we can actually use fake zeroed key */
+	if (params && params->integrity_key_size)
+		dmdi.u.integrity.vk = crypt_alloc_volume_key(params->integrity_key_size, NULL);
+
 	r = dm_create_device(cd, tmp_name, "INTEGRITY", &dmdi, 0);
+
+	crypt_free_volume_key(dmdi.u.integrity.vk);
 	if (r)
 		return r;
 
