@@ -1,8 +1,8 @@
 /*
  * utils_pbkdf - PBKDF ssettings for libcryptsetup
  *
- * Copyright (C) 2009-2017, Red Hat, Inc. All rights reserved.
- * Copyright (C) 2009-2017, Milan Broz
+ * Copyright (C) 2009-2018, Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2009-2018, Milan Broz
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -38,12 +38,12 @@ const struct crypt_pbkdf_type default_luks1 = {
 	.time_ms = DEFAULT_LUKS1_ITER_TIME
 };
 
-static uint32_t adjusted_pbkdf_memory(void)
+static uint32_t adjusted_phys_memory(void)
 {
 	uint64_t memory_kb = crypt_getphysmemory_kb();
 
 	/* Ignore bogus value */
-	if (memory_kb < (128 * 1024))
+	if (memory_kb < (128 * 1024) || memory_kb > UINT32_MAX)
 		return DEFAULT_LUKS2_MEMORY_KB;
 
 	/*
@@ -52,10 +52,7 @@ static uint32_t adjusted_pbkdf_memory(void)
 	 */
 	memory_kb /= 2;
 
-	if (memory_kb < DEFAULT_LUKS2_MEMORY_KB)
-		return (uint32_t)memory_kb;
-
-	return DEFAULT_LUKS2_MEMORY_KB;
+	return memory_kb;
 }
 
 /*
@@ -64,6 +61,7 @@ static uint32_t adjusted_pbkdf_memory(void)
 int verify_pbkdf_params(struct crypt_device *cd,
 			const struct crypt_pbkdf_type *pbkdf)
 {
+	struct crypt_pbkdf_limits pbkdf_limits;
 	const char *pbkdf_type;
 	int r = 0;
 
@@ -73,40 +71,64 @@ int verify_pbkdf_params(struct crypt_device *cd,
 	/* TODO: initialise crypto and check the hash and pbkdf are both available */
 	r = crypt_parse_pbkdf(pbkdf->type, &pbkdf_type);
 	if (r < 0) {
-		log_err(cd, _("Unknown PBKDF type %s.\n"), pbkdf->type);
+		log_err(cd, _("Unknown PBKDF type %s."), pbkdf->type);
 		return r;
 	}
 
+	r = crypt_pbkdf_get_limits(pbkdf->type, &pbkdf_limits);
+	if (r < 0)
+		return r;
+
 	if (crypt_get_type(cd) &&
 	    !strcmp(crypt_get_type(cd), CRYPT_LUKS1) &&
 	    strcmp(pbkdf_type, CRYPT_KDF_PBKDF2)) {
-		log_err(cd, _("Requested PBKDF type is not supported for LUKS1.\n"));
+		log_err(cd, _("Requested PBKDF type is not supported for LUKS1."));
 		return -EINVAL;
 	}
 
 	if (!strcmp(pbkdf_type, CRYPT_KDF_PBKDF2)) {
 		if (pbkdf->max_memory_kb || pbkdf->parallel_threads) {
-			log_err(cd, _("PBKDF max memory or parallel threads must not be set with pbkdf2.\n"));
+			log_err(cd, _("PBKDF max memory or parallel threads must not be set with pbkdf2."));
+			return -EINVAL;
+		}
+		if (pbkdf->flags & CRYPT_PBKDF_NO_BENCHMARK &&
+		    pbkdf->iterations < pbkdf_limits.min_iterations) {
+			log_err(cd, _("Forced iteration count is too low for %s (minimum is %u)."),
+				pbkdf_type, pbkdf_limits.min_iterations);
 			return -EINVAL;
 		}
 		return 0;
 	}
 
-	if (pbkdf->max_memory_kb > MAX_PBKDF_MEMORY) {
-		log_err(cd, _("Requested maximum PBKDF memory cost is too high (maximum is %d kilobytes).\n"),
-			MAX_PBKDF_MEMORY);
+	/* TODO: properly define minimal iterations and also minimal memory values */
+	if (pbkdf->flags & CRYPT_PBKDF_NO_BENCHMARK) {
+		if (pbkdf->iterations < pbkdf_limits.min_iterations) {
+			log_err(cd, _("Forced iteration count is too low for %s (minimum is %u)."),
+				pbkdf_type, pbkdf_limits.min_iterations);
+			r = -EINVAL;
+		}
+		if (pbkdf->max_memory_kb < pbkdf_limits.min_memory) {
+			log_err(cd, _("Forced memory cost is too low for %s (minimum is %u kilobytes)."),
+				pbkdf_type, pbkdf_limits.min_memory);
+			r = -EINVAL;
+		}
+	}
+
+	if (pbkdf->max_memory_kb > pbkdf_limits.max_memory) {
+		log_err(cd, _("Requested maximum PBKDF memory cost is too high (maximum is %d kilobytes)."),
+			pbkdf_limits.max_memory);
 		r = -EINVAL;
 	}
 	if (!pbkdf->max_memory_kb) {
-		log_err(cd, _("Requested maximum PBKDF memory can not be zero.\n"));
+		log_err(cd, _("Requested maximum PBKDF memory can not be zero."));
 		r = -EINVAL;
 	}
 	if (!pbkdf->parallel_threads) {
-		log_err(cd, _("Requested PBKDF parallel threads can not be zero.\n"));
+		log_err(cd, _("Requested PBKDF parallel threads can not be zero."));
 		r = -EINVAL;
 	}
 	if (!pbkdf->time_ms) {
-		log_err(cd, _("Requested PBKDF target time can not be zero.\n"));
+		log_err(cd, _("Requested PBKDF target time can not be zero."));
 		r = -EINVAL;
 	}
 
@@ -118,6 +140,7 @@ int init_pbkdf_type(struct crypt_device *cd,
 		    const char *dev_type)
 {
 	struct crypt_pbkdf_type *cd_pbkdf = crypt_get_pbkdf(cd);
+	struct crypt_pbkdf_limits pbkdf_limits;
 	const char *hash, *type;
 	unsigned cpus;
 	uint32_t old_flags, memory_kb;
@@ -132,6 +155,10 @@ int init_pbkdf_type(struct crypt_device *cd,
 	if (r)
 		return r;
 
+	r = crypt_pbkdf_get_limits(pbkdf->type, &pbkdf_limits);
+	if (r < 0)
+		return r;
+
 	/*
 	 * Crypto backend may be not initialized here,
 	 * cannot check if algorithms are really available.
@@ -168,10 +195,10 @@ int init_pbkdf_type(struct crypt_device *cd,
 	cd_pbkdf->max_memory_kb = pbkdf->max_memory_kb;
 	cd_pbkdf->parallel_threads = pbkdf->parallel_threads;
 
-	if (cd_pbkdf->parallel_threads > MAX_PBKDF_THREADS) {
+	if (cd_pbkdf->parallel_threads > pbkdf_limits.max_parallel) {
 		log_dbg("Maximum PBKDF threads is %d (requested %d).",
-			MAX_PBKDF_THREADS, cd_pbkdf->parallel_threads);
-		cd_pbkdf->parallel_threads = MAX_PBKDF_THREADS;
+			pbkdf_limits.max_parallel, cd_pbkdf->parallel_threads);
+		cd_pbkdf->parallel_threads = pbkdf_limits.max_parallel;
 	}
 
 	if (cd_pbkdf->parallel_threads) {
@@ -185,7 +212,7 @@ int init_pbkdf_type(struct crypt_device *cd,
 	}
 
 	if (cd_pbkdf->max_memory_kb) {
-		memory_kb = adjusted_pbkdf_memory();
+		memory_kb = adjusted_phys_memory();
 		if (cd_pbkdf->max_memory_kb > memory_kb) {
 			log_dbg("Not enough physical memory detected, "
 				"PBKDF max memory decreased from %dkB to %dkB.",
@@ -224,6 +251,19 @@ const struct crypt_pbkdf_type *crypt_get_pbkdf_type(struct crypt_device *cd)
 	return crypt_get_pbkdf(cd)->type ? crypt_get_pbkdf(cd) : NULL;
 }
 
+const struct crypt_pbkdf_type *crypt_get_pbkdf_default(const char *type)
+{
+	if (!type)
+		return NULL;
+
+	if (!strcmp(type, CRYPT_LUKS1))
+		return &default_luks1;
+	else if (!strcmp(type, CRYPT_LUKS2))
+		return &default_luks2;
+
+	return NULL;
+}
+
 void crypt_set_iteration_time(struct crypt_device *cd, uint64_t iteration_time_ms)
 {
 	struct crypt_pbkdf_type *pbkdf;
